import asyncio
import logging
import random
from redis import RedisError
import redis.asyncio

from .common import REDIS_KEYS


class ReconnectingRedis(redis.asyncio.Redis):
    def __init__(self, suffix, host, port, db, password, retry_on_timeout):
        self._log = logging.getLogger(__name__)
        self._retry_delay = 1
        self._name = "redis%s" % ("-%s" % suffix if suffix else "")
        self._host = host
        self._port = port
        self._pool = redis.asyncio.ConnectionPool(
            host=host,
            port=port,
            db=db,
            password=password,
            retry_on_timeout=retry_on_timeout,
        )
        super().__init__(connection_pool=self._pool)
        self.psub = self.pubsub()

    async def subscribe(self, callback, *channels):
        while True:
            try:
                await self.psub.subscribe(*channels)
                self._retry_delay = 1
                self._log.debug("%s connected", self._name)

                while True:
                    # Type check complains about iterating over a Coroutine type, but this is how
                    # the documentation indicates one should listen for messages.
                    async for msg in self.psub.listen():  # type: ignore
                        callback(msg)
                    await asyncio.sleep(0.01)

            except Exception as err:
                self._log.error(
                    "%s failed to connect to %s:%s - %s",
                    self._name,
                    self._host,
                    self._port,
                    err,
                )

                await self.psub.reset()

                self._retry_delay *= 2
                self._retry_delay += random.random()
                self._retry_delay = min(self._retry_delay, 10)

            await asyncio.sleep(self._retry_delay)



import cv2
import numpy

from flipdisc import util

__all__ = [
    "INTER_NEAREST",
    "INTER_LINEAR",
    "INTER_AREA",
    "INTER_CUBIC",
    "INTER_LANCZOS4",
    "resize_stretch",
    "resize_cover",
    "resize_contain",
    "resize_factor",
    "load_image",
]


# A nearest-neighbor interpolation.
INTER_NEAREST = cv2.INTER_NEAREST

# A bilinear interpolation.
INTER_LINEAR = cv2.INTER_LINEAR

# Resampling using pixel area relation.
# It may be a preferred method for image decimation,
# as it gives moire'-free results. But when the image is zoomed,
# it is similar to the INTER_NEAREST method.
INTER_AREA = cv2.INTER_AREA

# A bicubic interpolation over 4x4 pixel neighborhood.
INTER_CUBIC = cv2.INTER_CUBIC

# A Lanczos interpolation over 8x8 pixel neighborhood.
INTER_LANCZOS4 = cv2.INTER_LANCZOS4


def resize_stretch(im, out_size, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)

    Return the stretched image im that fills out_size.
    """
    return cv2.resize(im, out_size, interpolation=interpolation)


def resize_contain(im, out_size, bgcolor=0, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)
    * bgcolor: integer >= 0, used to fill the background
    * interpolation: interpolation method to use

    Return a scaled image im to the largest size that fits out_size.
    """
    im_height, im_width = im.shape[:2]
    im_ratio = im_width / float(im_height)
    out_ratio = out_size[0] / float(out_size[1])
    if im_ratio <= out_ratio:
        out_width = int(out_size[1] * im_ratio)
        out_height = out_size[1]
    else:
        out_width = out_size[0]
        out_height = int(out_size[0] / im_ratio)

    if 0 in (out_width, out_height):
        return numpy.zeros((out_size[1], out_size[0]), im.dtype)

    out = cv2.resize(im, (out_width, out_height), interpolation=interpolation)
    # If the output is too small, center it based on the specified size.
    if out.shape[1] < out_size[0]:
        extra_width = out_size[0] - out.shape[1]
        one_more = extra_width % 2
        out1 = numpy.zeros((out.shape[0], out_size[0]) + (out.shape[2:]), out.dtype)
        if bgcolor:
            out1.fill(bgcolor)
        out1[:, extra_width / 2 : -(extra_width / 2 + one_more)] = out
        out = out1
    if out.shape[0] < out_size[1]:
        extra_height = out_size[1] - out.shape[0]
        one_more = extra_height % 2
        out1 = numpy.zeros((out_size[1], out.shape[1]) + (out.shape[2:]), out.dtype)
        if bgcolor:
            out1.fill(bgcolor)
        out1[extra_height / 2 : -(extra_height / 2 + one_more), :] = out
        out = out1

    return out


def resize_factor(im, out, fx, fy, interpolation=None):
    """
    * im: numpy array
    * out: numpy array
    * fx: float x-axis scale factor
    * fy: float y-axis scale factor
    * interpolation: interpolation method to use

    Resize image and paste the result over the center of out; edges are
    cropped if necessary.
    """
    assert fx > 0, "fx must be positive"
    assert fy > 0, "fy must be positive"

    if interpolation is None:
        if fx < 1 or fy < 1:
            interpolation = cv2.INTER_AREA
        else:
            interpolation = cv2.INTER_NEAREST

    scaled_size = (int(fy * im.shape[1]), int(fx * im.shape[0]))
    if 0 in scaled_size:
        return out

    scaled = cv2.resize(im, scaled_size, interpolation=interpolation)
    ih, iw = scaled.shape[:2]
    oh, ow = out.shape[:2]

    x = y = ix = iy = 0
    x2, y2 = x + ow, y + oh
    ix2, iy2 = ix + iw, iy + ih

    if iw < ow:
        x = (ow - iw) / 2
        x2 = x + iw
    elif iw > ow:
        ix = (iw - ow) / 2
        ix2 = ix + ow

    if ih < oh:
        y = (oh - ih) / 2
        y2 = y + ih
    elif ih > oh:
        iy = (ih - oh) / 2
        iy2 = iy + oh

    out[y:y2, x:x2] = scaled[iy:iy2, ix:ix2]
    return out


def resize_cover(im, out_size, interpolation=INTER_NEAREST):
    """
    * im: numpy array
    * out_size: tuple of integers (width, height)
    * interpolation: interpolation method to use

    Return an image such that im covers out_size; edges are cropped if necessary.
    """
    im_height, im_width = im.shape[:2]
    im_ratio = im_width / float(im_height)
    out_ratio = out_size[0] / float(out_size[1])
    if im_ratio <= out_ratio:
        out_width = out_size[0]
        out_height = int(out_size[0] / im_ratio)
    else:
        out_width = int(out_size[1] * im_ratio)
        out_height = out_size[1]

    out = cv2.resize(im, (out_width, out_height), interpolation=interpolation)
    # If the output is too big, crop edges to match the specified size.
    if out.shape[1] > out_size[0]:
        extra_width = out.shape[1] - out_size[0]
        one_more = extra_width % 2
        out = out[:, extra_width / 2 : -(extra_width / 2 + one_more)]
    if out.shape[0] > out_size[1]:
        extra_height = out.shape[0] - out_size[1]
        one_more = extra_height % 2
        out = out[extra_height / 2 : -(extra_height / 2 + one_more), :]

    return out


def load_image(filename, area_size, binarize=0, padding=0):
    img = cv2.imread(filename)
    if img is None:
        return

    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    if padding:
        gray = util.pad(gray, padding)

    # Resize if needed.
    resize = max(gray.shape) > min(area_size)
    if resize:
        gray = util.resize(gray, area_size)

    # Binarize.
    if binarize is not None:
        if binarize >= 0:
            gray[gray <= binarize] = 0
            gray[gray > binarize] = 255
        else:
            # Use otsu.
            gray = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]

    # Copy image to the center of area.
    area = numpy.zeros(area_size, numpy.uint8)
    util.copy_to_center(area, gray)

    return area



import numpy as np
cimport cython
cimport numpy as np

DTYPE = np.float

ctypedef np.float_t DTYPE_t


@cython.boundscheck(False)
@cython.wraparound(False)
def dither_floydsteinberg(np.ndarray[DTYPE_t, ndim=2] im):
    """
          p   7
      3   5   1

    1/16
    """
    assert im.dtype == DTYPE

    cdef int rows = im.shape[0]
    cdef int cols = im.shape[1]
    cdef DTYPE_t old, error
    cdef int i, j

    for i in range(rows):
        for j in range(cols):
            old = im[i, j]
            im[i, j] = 255 if old > 127 else 0
            error = old - im[i, j]

            if j + 1 < cols:
                im[i    , j + 1] = im[i    , j + 1] + error * (7/16.)
            if i + 1 < rows:
                if j - 1 >= 0:
                    im[i + 1, j - 1] = im[i + 1, j - 1] + error * (3/16.)
                im[i + 1, j    ] = im[i + 1, j    ] + error * (5/16.)
                if j + 1 < cols:
                    im[i + 1, j + 1] = im[i + 1, j + 1] + error * (1/16.)


@cython.boundscheck(False)
@cython.wraparound(False)
def dither_atkinson(np.ndarray[DTYPE_t, ndim=2] im):
    """
          p   1   1
      1   1   1
          1

    1/8 (reduced color bleed)
    """
    assert im.dtype == DTYPE

    cdef int rows = im.shape[0]
    cdef int cols = im.shape[1]
    cdef DTYPE_t old, error
    cdef int i, j

    for i in range(rows):
        for j in range(cols):
            old = im[i, j]
            im[i, j] = 255 if old > 127 else 0
            error = (old - im[i, j]) / 8.

            if j + 1 < cols:
                im[i, j + 1] = im[i, j + 1] + error
                if j + 2 < cols:
                    im[i, j + 2] = im[i, j + 2] + error
            if i + 1 < rows:
                if j - 1 >= 0:
                    im[i + 1, j - 1] = im[i + 1, j - 1] + error
                im[i + 1, j] = im[i + 1, j] + error
                if j + 1 < cols:
                    im[i + 1, j + 1] = im[i + 1, j + 1] + error
                if i + 2 < rows:
                    im[i + 2, j] = im[i + 2, j] + error



import json
import struct
import logging
import logging.config
import asyncio
import numpy
import collections
import time
from redis import RedisError

import zmq
import zmq.asyncio

from .red import ReconnectingRedis
from .common import REDIS_KEYS, INPUT_STREAM, HDMI_INPUT_STREAM, OUTPUT_STREAM
from .common import ScheduledFunction

from typing import Any, Callable, Dict, List, Optional, Union

__all__ = ["Application"]


class Application(object):
    def __init__(
        self,
        name: str,
        config: Union[Dict[Any, Any], str],
        setup_input: bool = True,
        setup_output: bool = True,
        setup_hdmi_input: bool = True,
        verbose: bool = False,
    ) -> None:
        """
        An user app instance.

        * name should be a string. It will be used to store app data on redis
          and to identify the app through the http server.
          Accessible through instance.name

        * config can be either a string or a dict, if it's a string
          json.load(open(config)) will be used.
          + "logging" will be passed to logging.config.dictConfig(..)
          + If "output_stream", "input_stream" or "hdmi_stream" keys are present,
          they must specify at least a "width" and "height" subkeys.
          + If a "redis" key is present it will be used to configure the redis
          client.
          + The "settings" key must be used for holding settings specific
          to the app that can be updated while it's running.
          The app is free to use any other keys.
          Accessible through instance.config

        * verbose should be a boolean.
          Accessible through instance.verbose

        * if setup_input is True a SUB socket for the input stream will be
          configured for the topic framework.INPUT_STREAM using
          config["input_stream"]. The same applies for setup_hdmi_input
          with topic frame.HDMI_INPUT_STREAM and config["hdmi_stream"].

        * if setup_output is True a PUB socket for the output stream will be
          configured using config["output_stream"]
        """

        def load_config(config: Union[Dict[Any, Any], str]) -> Dict[Any, Any]:
            loaded_config: Dict[Any, Any]
            if not isinstance(config, dict):
                # Load config from json file.
                loaded_config = json.load(open(config))
            else:
                loaded_config = config
            if "logging" in loaded_config:
                logging.config.dictConfig(loaded_config["logging"])
            else:
                logging.basicConfig(
                    level=logging.DEBUG if verbose else logging.INFO,
                    format="%(asctime)s %(levelname)s %(name)s %(process)d %(message)s",
                )
            return loaded_config

        self.name = str.encode(name)
        self.verbose = verbose
        self.config: Dict[Any, Any] = load_config(config)
        self._log = logging.getLogger(__name__)

        self._loop = asyncio.get_event_loop()
        self._cb_app_heartbeat = None
        self._scheduled_functions = {}

        # List of redis channels to subscribe.
        self._sub_channels = []

        self._input_callback = {}
        self._loop = asyncio.get_event_loop()

        if "input_stream" in self.config:
            input_width = self.config["input_stream"]["width"]
            input_height = self.config["input_stream"]["height"]
            self._bgr_shape = (input_height, input_width, 3)
            self._depth_shape = (input_height, input_width)
            self._sub_channels.append(REDIS_KEYS.SYS_INPUT_CHANNEL.value)
        else:
            setup_input = False
            self._bgr_shape = None
            self._depth_shape = None

        if "hdmi_stream" in self.config:
            hdmi_width = self.config["hdmi_stream"]["width"]
            hdmi_height = self.config["hdmi_stream"]["height"]
            self._hdmi_shape = (hdmi_height, hdmi_width, 3)
            self._sub_channels.append(REDIS_KEYS.SYS_HDMI_CHANNEL.value)
        else:
            setup_hdmi_input = False
            self._hdmi_shape = None

        self._bgr_dtype = "uint8"
        self._depth_dtype = "uint16"
        self._hdmi_dtype = "uint8"

        if "output_stream" in self.config:
            output_width = self.config["output_stream"]["width"]
            output_height = self.config["output_stream"]["height"]
            self._bin_shape = (output_height, output_width)
            self._out_transition = self.config["output_stream"].get("transition", False)
            self._out_preview = self.config["output_stream"].get("preview", False)
            subtopic = self.config["output_stream"].get("subtopic")
            self._sub_channels.append(REDIS_KEYS.SYS_OUTPUT_CHANNEL.value)
        else:
            setup_output = False
            subtopic = None
            self._bin_shape = None
            self._out_transition = False
            self._out_preview = False
        self._bin_dtype = "uint8"
        if self._out_transition and not self.name:
            raise ValueError("App name is required to use transitions")

        self._app_rkey: Optional[bytes] = None
        if self.name:
            self._app_rkey = self.name
            if subtopic is not None:
                self._app_rkey += b":" + str.encode(subtopic)
            self._sub_channels.append(REDIS_KEYS.APP_CHANNEL.value + self._app_rkey)

        self._ctx = zmq.asyncio.Context()
        self._in_stream = {}
        self._sent: int = 0
        self._out_socket = None
        self._out_topic = None
        if setup_input:
            self.setup_input()
        if setup_hdmi_input:
            self.setup_input("hdmi_stream", HDMI_INPUT_STREAM)
        if setup_output:
            self.setup_output()

    async def setup_redis(self):
        self._cb_app_heartbeat = None
        self._redis_sub_callback = None
        self._red: Optional[ReconnectingRedis] = None
        self._red_sub: Optional[ReconnectingRedis] = None
        self._pubsub = None

        config = self.config.get("redis", {})

        if config.get("enabled", True) == False:
            return

        host = config.get("host", "localhost")
        port = config.get("port", 6379)
        db = config.get("db", None)
        pwd = config.get("auth", None)
        channels = self._sub_channels

        # Keep one redis connection for regular commands.
        self._red = ReconnectingRedis(
            suffix=None,
            host=host,
            port=port,
            db=db,
            password=pwd,
            retry_on_timeout=True,
        )

        # And another redis connection for pubsub.
        self._red_sub = ReconnectingRedis(
            suffix=None,
            host=host,
            port=port,
            db=db,
            password=pwd,
            retry_on_timeout=True,
        )

        if channels:
            self._loop.create_task(
                self._red_sub.subscribe(self._on_redis_message, *channels)
            )

    def setup_input(
        self,
        cfg: str = "input_stream",
        topic: str = INPUT_STREAM,
        bind: bool = False,
        watermark: int = 0,
    ) -> None:
        """
        Configure a ZMQ PUB socket for the input stream.
        """
        if cfg in self._in_stream:
            # Already configured the input stream, nothing to do.
            return

        sock_address = self.config[cfg]["socket"]
        in_socket = self._ctx.socket(zmq.SUB)

        in_socket.setsockopt_string(zmq.SUBSCRIBE, topic)
        if watermark > 0:
            in_socket.set_hwm(watermark)

        if bind:
            in_socket.bind(sock_address)
        else:
            in_socket.connect(sock_address)

        # TODO(wrigby): This is quite hacky - we shouldn't be doing this
        callback = getattr(self, "_input_callback_%s" % cfg, None)
        if callback is None:
            raise Exception("Implementation not available for %s configuration" % cfg)

        async def receive_stream(socket: zmq.asyncio.Socket, callback: Callable):
            while True:
                try:
                    msg = await socket.recv_multipart()
                    callback(msg)
                except Exception as e:
                    self._log.error("Error waiting for ZMQ stream message")
                    self._log.exception(e)

        self._in_stream[cfg] = self._loop.create_task(
            receive_stream(in_socket, callback)
        )

        self._log.debug(
            "SUB socket %s to %s, topic %s",
            "bound" if bind else "connected",
            sock_address,
            topic,
        )

    def set_input_callback(
        self, function: Callable, stream: str = "input_stream"
    ) -> None:
        """
        Define a callback to be invoked on messages received through
        the socket configured with setup_input.
        """
        self._input_callback[stream] = function

    def _input_callback_output_stream(self, msg: List[bytes]) -> None:
        cb = self._input_callback.get("output_stream")

        if cb is None or msg is None:
            return

        if len(msg) != 3:
            self._log.error("expected output message of size %d, got %d", 3, len(msg))
            return

        topic = msg[0]
        subtopic: Optional[bytes] = topic.split(b":", 1)[1].rstrip(b"\x00") if b":" in topic else None
        frame_num = struct.unpack("i", msg[1])[0]
        binimage_frame = msg[2]

        bin_image = numpy.frombuffer(binimage_frame, dtype=self._bin_dtype)
        try:
            bin_image = bin_image.reshape(self._bin_shape)
        except ValueError:
            self._log.exception("ignoring bad frame")
            return

        cb(app=self, subtopic=subtopic, frame_num=frame_num, bin_image=bin_image)

    # Preview stream is processed the same way as the output stream.
    _input_callback_preview_stream = _input_callback_output_stream

    def _input_callback_input_transition_stream(self, msg: List[bytes]):
        cb = self._input_callback.get("transition_stream")
        if cb is None:
            return

        if len(msg) != 4:
            self._log.error(
                "expected transition message of size %d, got %d", 4, len(msg)
            )
            return

        # topic = msg[0]  # unused.
        app_name = str(msg[1])[:-1]
        frame_num = struct.unpack("i", msg[2])[0]
        binimage_frame = msg[3]

        bin_image = numpy.frombuffer(binimage_frame, dtype=self._bin_dtype)
        try:
            bin_image = bin_image.reshape(self._bin_shape)
        except ValueError:
            self._log.exception("ignoring bad frame")
            return

        cb(app=self, frame_from=app_name, frame_num=frame_num, bin_image=bin_image)

    def _input_callback_input_stream(self, msg: List[bytes]):
        cb = self._input_callback.get("input_stream")
        if cb is None:
            return

        if len(msg) != 4:
            self._log.error("expected input message of size %d, got %d", 4, len(msg))
            return

        # topic = msg[0]  # unused.
        frame_num = struct.unpack("i", msg[1])[0]
        depth_frame = msg[2]
        bgr_frame = msg[3]

        depth = numpy.frombuffer(depth_frame, dtype=self._depth_dtype)
        bgr = numpy.frombuffer(bgr_frame, dtype=self._bgr_dtype)
        try:
            depth = depth.reshape(self._depth_shape)
            bgr = bgr.reshape(self._bgr_shape)
        except ValueError:
            self._log.exception("ignoring bad frame")
            return

        cb(app=self, frame_num=frame_num, depth=depth, bgr=bgr)

    def _input_callback_hdmi_stream(self, msg: List[bytes]):
        cb = self._input_callback.get("hdmi_stream")
        if cb is None:
            return

        if len(msg) != 3:
            self._log.error("expected hdmi message of size %d, got %d", 3, len(msg))
            return

        frame_num = struct.unpack("i", msg[1])[0]
        bgr_frame = msg[2]

        bgr = numpy.frombuffer(bgr_frame, dtype=self._hdmi_dtype)
        try:
            bgr = bgr.reshape(self._hdmi_shape)
        except ValueError:
            self._log.exception("ignoring bad frame")
            return

        cb(app=self, frame_num=frame_num, bgr=bgr)

    def setup_output(
        self, cfg: str = "output_stream", bind: Optional[bool] = None
    ) -> None:
        """
        Configure a ZMQ PUB socket for the output stream.
        """
        if self._out_socket:
            return

        sock_address = self.config[cfg]["socket"]
        subtopic: str = self.config[cfg].get("subtopic")
        subtopic = ":%s" % subtopic if subtopic else ""

        self._out_socket = self._ctx.socket(zmq.PUB)
        self._out_topic = bytes(OUTPUT_STREAM + subtopic, encoding="utf8")
        if bind is None:
            bind = (not self._out_transition) and (not self._out_preview)
        if bind:
            self._out_socket.bind(sock_address)
        else:
            self._out_socket.connect(sock_address)

        self._log.debug(
            "PUB socket %s to %s, topic %s",
            "bound" if bind else "connected",
            sock_address,
            self._out_topic,
        )

    async def send_output(
        self, result: numpy.ndarray, topic: Optional[bytes] = None
    ) -> int:
        """
        Publish result using the output socket configured by setup_output.

        An integer is returned. This number indicates the number of frames
        sent so far, i.e. the number of send_output invocations, starting
        from 1, wrapping around to 0 at UINT32_MAX.
        """
        if not self._out_socket:
            raise Exception("output stream not configured")

        # Wrap the sent count to uint32
        self._sent = (self._sent + 1) & 0xFFFFFFFF
        topic = self._out_topic if topic is None else bytes(topic)
        if self._out_transition:
            # When (potentially) using transitions, also send the app name.
            data = [
                b"%s\x00" % topic,
                b"%s\x00" % self.name,
                struct.pack("I", self._sent),
                result.tobytes(),
            ]
        else:
            data = [b"%s\x00" % topic, struct.pack("I", self._sent), result.tobytes()]
        await self._out_socket.send_multipart(data, copy=False)
        return self._sent

    async def notify(self, channel, data):
        """Send json.dumps(data) to a channel using redis pubsub."""
        if not self._red:
            raise Exception("redis not enabled")
        self._log.debug("Redis outgoing, channel: %s, data: %s", channel, data)
        return await self._red.pubsub.publish(channel, json.dumps(data))

    def set_redis_callback(self, function: Callable):
        """
        Define a callback to be invoked on messages received through
        the redis pubsub.
        """
        self._redis_sub_callback = function

    def add_periodic_callback(
        self, function: Callable, float_sec: float, start: bool = False
    ) -> int:
        """
        Schedule function to be called once each n seconds. The callback
        will receive this instance as its sole argument.

        The returned key can be used with stop_periodic_callback to stop
        the scheduling.
        """
        periodic = ScheduledFunction(
           float_sec, function, [self], periodic=True, loop=self._loop
        )
        key = id(periodic)
        self._scheduled_functions[key] = periodic
        if start:
            periodic.start()
        return key

    def stop_periodic_callback(self, key: str):
        """
        Stop a periodic callback created with add_periodic_callback.
        """
        periodic = self._scheduled_functions.pop(key)
        periodic.stop()

    async def call_later(self, float_sec: float, function: Callable):
        """
        Schedule function to be called after float_sec seconds have passed.

        The returned handle can be used with cancel_call_later to cancel the
        future call.
        """
        func = ScheduledFunction(float_sec, function, [self], False, self._loop)
        func.start()
        return func

    def cancel_call_later(self, handle):
        """
        Stop a scheduled call_later function from running.
        """
        handle.stop()

    def stop_ioloop(self):
        """
        Stop the event loop.
        Note: call the cleanup function after this one if the app is quitting.
        """
        self._loop.stop()

    async def run(self):
        """
        Start the registered callbacks and the event loop.
        If a name was specified, the app will be registered on
        redis if it's enabled.
        """
        if self.name and self._red is not None:
            key = self.add_periodic_callback(self._app_heartbeat, 1, True)
            self._cb_app_heartbeat = self._scheduled_functions[key]
        for pf in self._scheduled_functions.values():
            if not pf.started:
                pf.start()
        tasks = [x.get_task() for x in self._scheduled_functions.values()]
        await asyncio.gather(*tasks)

    async def cleanup(self) -> None:
        """
        Call this before quitting to stop registered callbacks and
        unregister the app from redis.
        """
        for pf in self._scheduled_functions.values():
            pf.stop()
        for t in self._in_stream.values():
            t.cancel()
        if self._cb_app_heartbeat:
            self._cb_app_heartbeat.stop()
        if self.name and self._red is not None:
            # Unregister app from redis.
            await self._red.hdel(REDIS_KEYS.APPS.value, self._app_rkey) # type: ignore
            if self._red is not None:
                await self._red.close()
            if self._red_sub is not None:
                await self._red_sub.close()

    async def _app_heartbeat(self, app: "Application") -> None:
        # Update register on redis to indicate that the app is running well.
        self.config["timestamp"] = time.time()
        try:
            await self._red.hset( # type: ignore
                REDIS_KEYS.APPS.value, self._app_rkey, json.dumps(self.config) # type: ignore
            )
        except:
            self._log.error("No redis configuration was found")

    def _on_redis_message(self, msg):
        if msg is None:
            return
        msg_type = msg["type"]
        msg_channel = msg["channel"]
        content = msg["data"]

        self._log.debug(
            "Redis incoming, type: %s, channel: %s, msg: %s",
            msg_type,
            msg_channel,
            content,
        )

        if msg_type != "message":
            return

        data = json.loads(content)
        if msg_channel.startswith(REDIS_KEYS.APP_CHANNEL.value.decode()):
            # Update app settings.
            _update_settings(self.config["settings"], data)
            await self._app_heartbeat()
        elif msg_channel == REDIS_KEYS.SYS_INPUT_CHANNEL.value.decode():
            # Input stream update.
            _update_settings(self.config["input_stream"], data)
            input_height = self.config["input_stream"]["height"]
            input_width = self.config["input_stream"]["width"]
            self._bgr_shape = (input_height, input_width, 3)
            self._depth_shape = (input_height, input_width)
        elif msg_channel == REDIS_KEYS.SYS_OUTPUT_CHANNEL.value.decode():
            # Output stream update.
            _update_settings(self.config["output_stream"], data)
            self._bin_shape = (
                self.config["output_stream"]["height"],
                self.config["output_stream"]["width"],
            )

        if self._redis_sub_callback:
            self._log.debug("Redis sub callback")
            self._redis_sub_callback(app=self, channel=msg_channel, update=data)


def _update_settings(settings, new):
    # Restrict key updates to those that already exist and
    # enforce the values to be of same type.
    def update(orig_data, new_data):
        for key, new_value in new_data.items():
            if key in orig_data:
                if isinstance(new_value, collections.Mapping):
                    # Recurse on nested settings.
                    res = update(orig_data[key], new_value)
                    orig_data[key] = res
                else:
                    orig_data[key] = type(orig_data[key])(new_value)
        return orig_data

    update(settings, new)



import os
from setuptools import setup, Extension


def get_numpy_include():
    """Helper for lazy-importing numpy to get includes"""
    import numpy

    return numpy.get_include()


setup(
    name="flipdisc",
    version="1.0.1",
    url="https://github.com/breakfastny/flipdisc-python-framework",
    packages=["flipdisc", "flipdisc.framework"],
    package_data={
        "flipdisc.framework": ["*.json"],
    },
    python_requires='>=3.12',
    setup_requires=[
        "cffi<2",
        "cython",
        "numpy<2"
    ],
    cffi_modules=["flipdisc/build_particle.py:ffibuilder"],
    install_requires=[
        "cffi<2",
        "numpy<2",
        "jsonschema<5",
        "pyzmq<28",
        "redis<6"
    ],
    ext_modules=[
        Extension(
            "flipdisc.binarize",
            ["flipdisc/binarize.pyx"],
            extra_compile_args=["-g0", "-O2", "-march=native"],
            include_dirs=(f() for f in [get_numpy_include]),
        ),
    ],
    zip_safe=False,
)



import enum
import asyncio
from typing import Callable, Iterable, Optional


class REDIS_KEYS(enum.Enum):
    # hashtable used to store all active apps.
    # key is an app name and the value is a json.dumps of all its settings.
    APPS = b"fd:apps"
    # channel name used to send messages to specific apps.
    # When subscribing, use it as APP_CHANNEL + appname.
    APP_CHANNEL = b"fd:app:"
    # channel name used to send messages to all apps that use input settings.
    SYS_INPUT_CHANNEL = b"fd:system:input"
    # channel name used to send messages to all apps that use hdmi settings.
    SYS_HDMI_CHANNEL = b"fd:system:hdmi"
    # channel name used to send messages to all apps that use output settings.
    SYS_OUTPUT_CHANNEL = b"fd:system:output"


INPUT_STREAM = "IN_STREAM"
HDMI_INPUT_STREAM = "HDMI_STREAM"

OUTPUT_STREAM = "OUT_STREAM"


class ScheduledFunction:
    def __init__(
        self,
        delay: float,
        function: Callable,
        args: Iterable,
        periodic: bool = False,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.delay = delay
        self.function = function
        self.periodic = periodic
        self.started = False
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._args = args
        self._task: Optional[asyncio.Task] = None

    def get_task(self):
        return self._task

    def start(self):
        if not self.started:
            self.started = True
            try:
                self._task = self._loop.create_task(self._run())
            except asyncio.CancelledError:
                pass

    def stop(self):
        if self.started and self._task is not None:
            self.started = False
            self._task.cancel()

    async def _run(self):
        if self.periodic == True:
            while True:
                await asyncio.gather(
                    asyncio.sleep(self.delay),
                    self.function(*self._args),
                )
        else:
            self._loop.call_later(self.delay, self.function)



import os

from cffi import FFI

FDL_PATHS = []
if "FDL_INCLUDE_DIR" in os.environ:
    FDL_PATHS.append(os.environ["FDL_INCLUDE_DIR"])
else:
    FDL_PATHS.append("/usr/local/include/fdl")
    FDL_PATHS.append("/usr/include/fdl")


def search_particle_h():
    for p in FDL_PATHS:
        check = os.path.join(p, "particle.h")
        if os.path.exists(p) and os.path.exists(check):
            return check

    raise SystemExit("** ERROR: particle.h not found at %s" % (",".join(FDL_PATHS)))


ffibuilder = FFI()
ffibuilder.set_source(
    "flipdisc._particle", '#include "fdl/particle.h"', libraries=["fparticle"],
    extra_compile_args=["-march=native"]
)

raw_cdef = open(search_particle_h()).read()
# Skip includes and other things. This assumes the first definition
# in particle.h starts with a struct.
raw_start = raw_cdef.find("struct ")
raw_end = raw_cdef.find("#ifdef __cplusplus", raw_start + 1)
ffibuilder.cdef(raw_cdef[raw_start:raw_end])

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)
